@@ -38,6 +38,7 @@ from zc.buildout.configparser import (
     section_header,
 )
 from . import recipes
+from . import jinja
 
 from pygls.server import LanguageServer
 from pygls.types import Position, Range, Location
@@ -48,6 +49,9 @@ logger = logging.getLogger(__name__)
 # We also tolerate ${section: without option or the ending } to generate completions.
 option_reference_re = re.compile(
     r'\${(?P<section>[-a-zA-Z0-9 ._]*):(?P<option>[-a-zA-Z0-9 ._]*)')
+# In this version, we don't tolerate the missing }
+option_reference_strict_re = re.compile(
+    r'\${(?P<section>[-a-zA-Z0-9 ._]*):(?P<option>[-a-zA-Z0-9 ._]*)}')
 
 # Matches of an unterminated ${section:
 section_reference_re = re.compile(r'.*\$\{(?P<section>[-a-zA-Z0-9 ._]*)[^:]*$')
@@ -110,7 +114,7 @@ class BuildoutOptionDefinition:
     self.value = value
     self.implicit_option = implicit_option
 
-  def __repr__(self):
+  def __repr__(self) -> str:
     locations = ' '.join(
         ['{} {}'.format(l.uri, l.range) for l in self.locations])
     return '{} ({})'.format(self.value, locations)
@@ -182,6 +186,7 @@ class Symbol:
       current_option_name: Optional[str] = None,
       referenced_section_name: Optional[str] = None,
       referenced_option_name: Optional[str] = None,
+      is_same_section_reference: bool = False,
   ):
     self._buildout = buildout
     self.kind = kind
@@ -190,8 +195,9 @@ class Symbol:
     self.current_option_name = current_option_name
     self.referenced_section_name = referenced_section_name
     self.referenced_option_name = referenced_option_name
+    self.is_same_section_reference = is_same_section_reference
 
-  def __repr__(self):
+  def __repr__(self) -> str:
     referenced = ""
     if self.referenced_section_name:
       referenced = f"referenced=${{{self.referenced_section_name}:{self.referenced_option_name}}}"
@@ -320,6 +326,7 @@ class BuildoutTemplate:
               current_option_name=current_option_name,
               referenced_section_name=referenced_section_name or
               current_section_name,
+              is_same_section_reference=referenced_section_name == '',
               referenced_option_name=referenced_option_name,
           )
         else:
@@ -349,6 +356,7 @@ class BuildoutTemplate:
               current_option_name=current_option_name,
               referenced_section_name=referenced_section_name or
               current_section_name,
+              is_same_section_reference=referenced_section_name == '',
           )
         else:
           logger.debug("section_reference_match was not in range, advancing")
@@ -384,6 +392,7 @@ class BuildoutTemplate:
             value=match.string[slice(*match.span())],
             referenced_section_name=match.group('section'),
             referenced_option_name=match.group('option'),
+            is_same_section_reference=match.group('section') == '',
         )
         symbol.section_range = Range(
             Position(
@@ -433,7 +442,6 @@ class BuildoutProfile(Dict[str, BuildoutSection], BuildoutTemplate):
     One exception is for profiles names software.cfg or buildout.cfg - we just assume
     that the template is valid for these profiles. For other profiles, we check if
     the profile really uses this template.
-    XXX this might not be good, but we don't resolve subsitutions inside buildout now.
     """
 
     # uri can be passed as relative or absolute. Let's build a set of absolute
@@ -457,72 +465,62 @@ class BuildoutProfile(Dict[str, BuildoutSection], BuildoutTemplate):
     if not os.path.exists(document.path):
       return None
 
-    for section_value in self.values():
+    for section_name, section_value in self.items():
       recipe = section_value.getRecipe()
       if recipe is not None:
-
         for template_option_name in recipe.template_options:
           template_option_value = section_value.get(template_option_name)
-          if (template_option_value is not None and
-              template_option_value.value in uris or
-              self.uri.endswith('/buildout.cfg') or
-              self.uri.endswith('/software.cfg')):
+          if template_option_value is not None:
+            template_option_value_uri = template_option_value.value
+            # expand substitutions
+            if '${' in template_option_value_uri:
+              section_value_dict = dict(section_value)
+              if '<' in section_value:
+                section_value_dict = dict(self[section_value['<'].value])
+                section_value_dict.update(dict(section_value))
+              logger.debug("We have a section reference %s in %s",
+                           template_option_value_uri, section_value_dict)
 
-            if slapos_instance_profile_filename_re.match(uri):
-              # a slapos "buildout profile as a template"
-              slapos_instance_profile = await open(
-                  ls,
-                  uri,
-                  allow_errors=True,
-                  open_as_buildout_profile=True,
+              def expand_section_reference(match: Match[str]) -> str:
+                referenced_section_name = match.group('section') or section_name
+                if referenced_section_name in self:
+                  referenced_section = self[referenced_section_name]
+                  if match.group('option') in referenced_section:
+                    return referenced_section[match.group('option')].value
+                return '\0'  # won't likely match a filename
+
+              template_option_value_uri = option_reference_strict_re.sub(
+                  expand_section_reference,
+                  template_option_value_uri,
               )
-              assert isinstance(slapos_instance_profile, BuildoutProfile)
-              slapos_instance_profile.second_level_buildout = slapos_instance_profile
-              slapos_instance_profile.buildout = self
+              logger.debug("Section reference expanded to: %s",
+                           template_option_value_uri)
 
-              # Add slapos instance generated sections.
-              slapos_instance_profile.section_header_locations.setdefault(
-                  'slap-connection',
-                  Location(uri='', range=Range(Position(), Position())))
-              slap_connection = BuildoutSection()
-              for k in (
-                  'computer-id',
-                  'partition-id',
-                  'server-url',
-                  'software-release-url',
-              ):
-                slap_connection[k] = BuildoutOptionDefinition(
-                    locations=[],
-                    value='',
-                    implicit_option=True,
+            # Normalize URI path, in case it contain double slashes, ./ or ..
+            template_option_value_parsed = urllib.parse.urlparse(
+                template_option_value_uri)
+            template_option_value_uri = urllib.parse.urlunparse(
+                template_option_value_parsed._replace(
+                    path=os.path.normpath(template_option_value_parsed.path)))
+
+            if template_option_value_uri in uris:
+              if slapos_instance_profile_filename_re.match(uri):
+                # a slapos "buildout profile as a template"
+                slapos_instance_profile = await open(
+                    ls,
+                    uri,
+                    allow_errors=True,
+                    force_open_as_buildout_profile=True,
                 )
-              slapos_instance_profile.setdefault(
-                  'slap-connection',
-                  slap_connection,
+                assert isinstance(slapos_instance_profile, BuildoutProfile)
+                slapos_instance_profile.second_level_buildout = slapos_instance_profile
+                slapos_instance_profile.buildout = self
+                return slapos_instance_profile
+              return BuildoutTemplate(
+                  uri=uri,
+                  source=document.source,
+                  buildout=self,
               )
-              slapos_instance_profile.section_header_locations.setdefault(
-                  'slap-network-information',
-                  Location(uri='', range=Range(Position(), Position())))
-              slap_network_information = BuildoutSection()
-              for k in ('local-ipv4', 'global-ipv6', 'network-interface',
-                        'tap-ipv4', 'tap-gateway', 'tap-netmask', 'tap-network',
-                        'global-ipv4-network'):
-                slap_network_information[k] = BuildoutOptionDefinition(
-                    locations=[],
-                    value='',
-                    implicit_option=True,
-                )
-              slapos_instance_profile.setdefault(
-                  'slap-network-information',
-                  slap_network_information,
-              )
-
-              return slapos_instance_profile
-            return BuildoutTemplate(
-                uri=uri,
-                source=document.source,
-                buildout=self,
-            )
     return None
 
   async def getSymbolAtPosition(self, position: Position) -> Optional[Symbol]:
@@ -573,7 +571,7 @@ class BuildoutProfile(Dict[str, BuildoutSection], BuildoutTemplate):
       if line_offset == 0:
         # we can be in the following cases (> denotes beginning of lines)
         # - a section header
-        #   [section]
+        #   >[section]
         # - a single line option and value:
         #   >option = value
         # - an option without value:
@@ -692,50 +690,59 @@ class BuildoutProfile(Dict[str, BuildoutSection], BuildoutTemplate):
     """
     option: BuildoutOptionDefinition
     option = self[section_name][option_name]
-    start_line = option.locations[-1].range.start.line
-    lines = self.source.splitlines(
-    )[start_line:option.locations[-1].range.end.line + 1]
-    is_multi_line_option = len(lines) > 1
-    for line_offset, option_value_text in enumerate(lines):
-      if option_value_text and option_value_text[0] not in '#;':
-        start_character = 0
+    location = option.locations[-1]
+    if location.uri == self.uri:
+      start_line = location.range.start.line
+      lines = self.source.splitlines()[start_line:location.range.end.line + 1]
+      is_multi_line_option = len(lines) > 1
+      for line_offset, option_value_text in enumerate(lines):
+        if option_value_text and option_value_text[0] not in '#;':
+          start_character = 0
 
-        if option_value_text.startswith(option_name):
-          option_name_text, option_value_text = option_value_text.split('=', 1)
-          start_character += len(option_name_text) + 1
+          if option_value_text.startswith(option_name):
+            option_name_text, option_value_text = option_value_text.split(
+                '=', 1)
+            start_character += len(option_name_text) + 1
 
-        start_character += len(option_value_text) - len(
-            option_value_text.lstrip())
-        option_value_text = option_value_text.strip()
-        if option_value_text:
-          if is_multi_line_option:
-            yield (
-                option_value_text,
-                Range(
-                    Position(
-                        start_line + line_offset,
-                        start_character,
-                    ),
-                    Position(
-                        start_line + line_offset,
-                        start_character + len(option_value_text),
-                    ),
-                ),
-            )
-          else:
-            for match in re.finditer(r'(\w+)', option_value_text):
-              yield (match.group(),
-                     Range(
-                         Position(
-                             start_line + line_offset,
-                             start_character + match.start(),
-                         ),
-                         Position(
-                             start_line + line_offset,
-                             start_character + match.start() +
-                             len(match.group()),
-                         ),
-                     ))
+          start_character += len(option_value_text) - len(
+              option_value_text.lstrip())
+          option_value_text = option_value_text.strip()
+          if option_value_text:
+            if is_multi_line_option:
+              yield (
+                  option_value_text,
+                  Range(
+                      Position(
+                          start_line + line_offset,
+                          start_character,
+                      ),
+                      Position(
+                          start_line + line_offset,
+                          start_character + len(option_value_text),
+                      ),
+                  ),
+              )
+            else:
+              for match in re.finditer(r'([^\s]+)', option_value_text):
+                yield (match.group(),
+                       Range(
+                           Position(
+                               start_line + line_offset,
+                               start_character + match.start(),
+                           ),
+                           Position(
+                               start_line + line_offset,
+                               start_character + match.start() +
+                               len(match.group()),
+                           ),
+                       ))
+
+  @staticmethod
+  def looksLikeBuildoutProfile(uri: URI) -> bool:
+    """Check if this URI looks like a buildout profile URI.
+    """
+    return (uri.endswith('.cfg') or uri.endswith('.cfg.in') or
+            uri.endswith('.cfg.j2') or uri.endswith('.cfg.jinja2'))
 
 
 class ResolvedBuildout(BuildoutProfile):
@@ -754,7 +761,7 @@ _resolved_buildout_cache: Dict[URI, ResolvedBuildout] = {}
 _extends_dependency_graph: Dict[URI, Set[URI]] = collections.defaultdict(set)
 
 
-def clearCache(uri: URI):
+def clearCache(uri: URI) -> None:
   """Clear all caches for uri.
 
   This is to be called when the document is modified.
@@ -764,7 +771,7 @@ def clearCache(uri: URI):
   _clearExtendCache(uri, set())
 
 
-def _clearExtendCache(uri: URI, done: Set[URI]):
+def _clearExtendCache(uri: URI, done: Set[URI]) -> None:
   """Clear the `extends` cache for URI.
 
   This is to be called for all URIs extended by `uri`.
@@ -792,7 +799,7 @@ _isurl = re.compile('([a-zA-Z0-9+.-]+)://').match
 async def parse(
     ls: LanguageServer,
     uri: URI,
-    allow_errors=True,
+    allow_errors: bool = True,
 ) -> BuildoutProfile:
   """
   Parse a sectioned setup file and return a non-resolved buildout.
@@ -823,7 +830,7 @@ async def parse(
 async def _parse(
     fp: TextIO,
     uri: URI,
-    allow_errors,
+    allow_errors: bool,
 ) -> BuildoutProfile:
   """Parse a sectioned setup file and return a non-resolved buildout.
 
@@ -848,8 +855,12 @@ async def _parse(
   # buildout default values
   sections['buildout'] = BuildoutSection()
   for k, v in _buildout_default_options.items():
+    if isinstance(v, tuple):
+      value = v[0]  # buildout < 2.9.3
+    else:
+      value = v.value
     sections['buildout'][k] = BuildoutOptionDefinition(
-        value=v.value,
+        value=value,
         locations=[Location(uri=uri, range=Range(Position(0), Position(0)))],
         implicit_option=True,
     )
@@ -858,13 +869,52 @@ async def _parse(
       locations=[Location(uri=uri, range=Range(Position(0), Position(0)))],
       implicit_option=True,
   )
-
   sections.section_header_locations['buildout'] = Location(
       uri="",
       range=Range(Position(), Position()),
   )
+  if slapos_instance_profile_filename_re.match(uri):
+    # Add slapos instance generated sections.
+    sections.section_header_locations.setdefault(
+        'slap-connection',
+        Location(uri='', range=Range(Position(), Position())))
+    slap_connection = BuildoutSection()
+    for k in (
+        'computer-id',
+        'partition-id',
+        'server-url',
+        'key-file',
+        'cert-file',
+        'software-release-url',
+    ):
+      slap_connection[k] = BuildoutOptionDefinition(
+          locations=[],
+          value='',
+          implicit_option=True,
+      )
+    sections.setdefault('slap-connection', slap_connection)
+    sections.section_header_locations.setdefault(
+        'slap-network-information',
+        Location(uri='', range=Range(Position(), Position())))
+    slap_network_information = BuildoutSection()
+    for k in (
+        'local-ipv4',
+        'global-ipv6',
+        'network-interface',
+        'tap-ipv4',
+        'tap-gateway',
+        'tap-netmask',
+        'tap-network',
+        'global-ipv4-network',
+    ):
+      slap_network_information[k] = BuildoutOptionDefinition(
+          locations=[],
+          value='',
+          implicit_option=True,
+      )
+    sections.setdefault('slap-network-information', slap_network_information)
 
-  in_jinja_context = False
+  jinja_parser = jinja.JinjaParser()
   cursect: Optional[Dict[str, BuildoutOptionDefinition]] = None
   blockmode = False
   optname: Optional[str] = None
@@ -874,17 +924,15 @@ async def _parse(
     line = fp.readline()
     if not line:
       break  # EOF
-
     lineno = lineno + 1
 
+    jinja_parser.feed(line)
+    if jinja_parser.is_in_jinja:
+      continue
+
     if line[0] in '#;':
       continue  # comment
 
-    if line.startswith('{') or in_jinja_context:
-      in_jinja_context = '}' in line
-    if in_jinja_context:
-      continue
-
     if line[0].isspace() and (cursect is not None) and optname:
       _line = line
       # continuation line
@@ -924,13 +972,16 @@ async def _parse(
               value=sectname,
               implicit_option=True,
           )
-          # this is a slapos.buildout extension
+          # _profile_base_location_ is a slapos.buildout extension
+          base_location = '.'
+          if '/' in uri:
+            base_location = uri[:uri.rfind('/')] + '/'
           cursect['_profile_base_location_'] = BuildoutOptionDefinition(
               locations=[
                   Location(
                       uri=uri, range=Range(Position(0, 0), Position(0, 0)))
               ],
-              value=uri,
+              value=base_location,
               implicit_option=True,
           )
 
@@ -999,33 +1050,47 @@ async def open(
     ls: LanguageServer,
     uri: URI,
     allow_errors: bool = True,
-    open_as_buildout_profile: bool = False,
+    force_open_as_buildout_profile: bool = False,
 ) -> Optional[Union[BuildoutTemplate, ResolvedBuildout]]:
   """Open an URI and returnes either a buildout or a profile connected to buildout.
 
-  In the case of slapos buildout templates (instance.cfg.in), it is both.
+  In the case of slapos buildout templates (instance.cfg.in), it is both. This is
+  not true for slapos buildout templates as jinja templates, which have their own
+  namespace as ${} and not as $${}.
+
+  force_open_as_buildout_profile is used to force assuming that this file is a
+  buildout profile (and not a buildout template).
 
   For buildout, it is a wrapper over _open which uses language server's workspace
   """
   document = ls.workspace.get_document(uri)
   logger.debug("open %s", uri)
-  if not open_as_buildout_profile:
+  if not force_open_as_buildout_profile:
+
+    def getCandidateBuildoutProfiles() -> Iterator[pathlib.Path]:
+      path = pathlib.Path(document.path).parent
+      for _ in range(3):  # look for buildouts up to 3 levels
+        # we sort just to have stable behavior
+        for profile in sorted(path.glob('*.cfg')):
+          yield profile
+        path = path.parent
+
     # First, try to read as a template, because buildout profiles can be templates.
     if slapos_instance_profile_filename_re.match(
         uri) or not uri.endswith('.cfg'):
-      logger.debug('Trying to open %s as a template', uri)
-      for buildout_path in sorted(
-          pathlib.Path(document.path).parent.glob('*.cfg'),
-          # we try buildout.cfg and software.cfg at last, because getTemplate
-          # returns a valid template for any file.
-          key=lambda path: path.name in ('buildout.cfg', 'software.cfg')):
-
-        # We don't use  buildout_path.resolve().as_uri(), because we have fake uri -> path mapping in tests
-        buildout_uri = str(buildout_path.resolve()).replace(
-            ls.workspace.root_path,
-            ls.workspace.root_uri,
-            1,
-        )
+      for buildout_path in getCandidateBuildoutProfiles():
+        resolved_path = str(buildout_path.resolve())
+        # For paths in workspace, we don't use buildout_path.resolve().as_uri(),
+        # because we have fake uri -> path mapping in tests
+        if resolved_path.startswith(ls.workspace.root_path):
+          buildout_uri = resolved_path.replace(
+              ls.workspace.root_path,
+              ls.workspace.root_uri,
+              1,
+          )
+        else:
+          # but we still need to support the case where the path is outside the workspace
+          buildout_uri = buildout_path.resolve().as_uri()
         logger.debug("Trying to find templates's buildout with %s -> %s",
                      buildout_path, buildout_uri)
         buildout = await _open(
@@ -1035,13 +1100,13 @@ async def open(
             [],
             allow_errors=allow_errors,
         )
-
         assert isinstance(buildout, BuildoutProfile)
         template = await buildout.getTemplate(ls, uri)
         if template is not None:
           return template
 
-  if uri.endswith('.cfg') or open_as_buildout_profile:
+  if BuildoutProfile.looksLikeBuildoutProfile(
+      uri) or force_open_as_buildout_profile:
     fp = io.StringIO(document.source)
     return await _open(ls, '', uri, [], allow_errors=allow_errors)
 
