@@ -1,12 +1,14 @@
+import asyncio
 import contextlib
 import io
+import itertools
 import logging
 import os
+import pathlib
 import re
 import tempfile
 import urllib.parse
-import pathlib
-from typing import Any, List, Union, Tuple, Iterable, Optional, Sequence
+from typing import Any, Iterable, List, Optional, Sequence, Set, Tuple, Union
 
 from zc.buildout import configparser
 from zc.buildout.buildout import Buildout
@@ -15,8 +17,8 @@ from zc.buildout.configparser import MissingSectionHeaderError, ParsingError
 from pygls.features import (
     COMPLETION,
     DEFINITION,
-    DOCUMENT_SYMBOL,
     DOCUMENT_LINK,
+    DOCUMENT_SYMBOL,
     HOVER,
     REFERENCES,
     TEXT_DOCUMENT_DID_CHANGE,
@@ -36,6 +38,8 @@ from pygls.types import (
     DidChangeWatchedFiles,
     DidOpenTextDocumentParams,
     DidSaveTextDocumentParams,
+    DocumentLink,
+    DocumentLinkParams,
     DocumentSymbol,
     DocumentSymbolParams,
     Hover,
@@ -47,15 +51,12 @@ from pygls.types import (
     Range,
     SymbolKind,
     TextDocumentPositionParams,
-    DocumentLink,
-    DocumentLinkParams,
     TextEdit,
     WorkspaceEdit,
 )
+from pygls.workspace import Document
 
-from . import buildout
-from . import recipes
-import itertools
+from . import buildout, jinja, recipes
 
 server = LanguageServer()
 
@@ -65,24 +66,30 @@ reference_re = re.compile(
 
 logger = logging.getLogger(__name__)
 
+DEBOUNCE_DELAY = 0.3
+
 
 def getOptionValue(
     option: Union[buildout.BuildoutOptionDefinition, str]) -> str:
   # Options read with our patch remember the position and have their values in
   # .value but options added by buildout for default values does not.
   # We normalize this here.
-  return getattr(option, 'value', option)
+  if isinstance(option, str):
+    return option
+  return option.value
 
 
 async def parseAndSendDiagnostics(
     ls: LanguageServer,
     uri: str,
-):
+) -> None:
+  await asyncio.sleep(DEBOUNCE_DELAY)
   diagnostics: List[Diagnostic] = []
   parsed = None
 
-  if (uri.endswith('.cfg') or uri.endswith('.cfg.in') or
-      uri.endswith('.cfg.j2')):
+  looks_like_profile = buildout.BuildoutProfile.looksLikeBuildoutProfile(uri)
+
+  if looks_like_profile:
     # parse errors
     try:
       parsed = await buildout.parse(
@@ -94,8 +101,7 @@ async def parseAndSendDiagnostics(
       if e.filename != uri:
         logger.debug("skipping error in external file %s", e.filename)
       elif isinstance(e, MissingSectionHeaderError):
-        if (uri.endswith('.cfg') or uri.endswith('.cfg.in') or
-            uri.endswith('.cfg.j2')):
+        if looks_like_profile:
           diagnostics.append(
               Diagnostic(
                   message=e.message,
@@ -106,8 +112,7 @@ async def parseAndSendDiagnostics(
                   source='buildout',
               ))
       else:
-        if (uri.endswith('.cfg') or uri.endswith('.cfg.in') or
-            uri.endswith('.cfg.j2')):
+        if looks_like_profile:
           for (lineno, _), msg in zip(e.errors, e.message.splitlines()[1:]):
             msg = msg.split(':', 1)[1].strip()
             diagnostics.append(
@@ -125,6 +130,13 @@ async def parseAndSendDiagnostics(
       uri=uri,
   )
   assert resolved_buildout is not None
+
+  installed_parts: Set[str] = set([])
+  if isinstance(resolved_buildout, buildout.BuildoutProfile):
+    if 'parts' in resolved_buildout['buildout']:
+      installed_parts = set((
+          v[0] for v in resolved_buildout.getOptionValues('buildout', 'parts')))
+
   async for symbol in resolved_buildout.getAllOptionReferenceSymbols():
     if symbol.referenced_section is None:
       diagnostics.append(
@@ -143,6 +155,9 @@ async def parseAndSendDiagnostics(
               symbol.referenced_option_name in
               symbol.referenced_section_recipe.generated_options):
         continue
+      # if a section is a macro, it's OK to self reference ${:missing}
+      if symbol.is_same_section_reference and symbol.current_section_name not in installed_parts:
+        continue
       diagnostics.append(
           Diagnostic(
               message=f'Option `{symbol.referenced_option_name}` does not exist in `{symbol.referenced_section_name}`.',
@@ -153,7 +168,8 @@ async def parseAndSendDiagnostics(
 
   if isinstance(resolved_buildout, buildout.BuildoutProfile):
     for section_name, section in resolved_buildout.items():
-      if resolved_buildout.section_header_locations[section_name].uri == uri:
+      if section_name in installed_parts and resolved_buildout.section_header_locations[
+          section_name].uri == uri:
         # check for required options
         recipe = section.getRecipe()
         if recipe:
@@ -172,12 +188,16 @@ async def parseAndSendDiagnostics(
                 ),)
 
     if 'parts' in resolved_buildout['buildout']:
+      jinja_parser = jinja.JinjaParser()
       for part_name, part_range in resolved_buildout.getOptionValues(
           'buildout', 'parts'):
         if part_name:
-          if part_name.startswith('${') or part_name.startswith('{{'):
-            # Assume buildout/jinja substitutions are OK.
-            continue
+          if part_name.startswith('${'):
+            continue  # assume substitutions are OK
+          jinja_parser.feed(part_name)
+          if jinja_parser.is_in_jinja:
+            continue  # ignore anything in jinja context
+
           if part_name not in resolved_buildout:
             diagnostics.append(
                 Diagnostic(
@@ -232,6 +252,7 @@ async def lsp_symbols(
     ls: LanguageServer,
     params: DocumentSymbolParams,
 ) -> List[DocumentSymbol]:
+  await asyncio.sleep(DEBOUNCE_DELAY)
   symbols: List[DocumentSymbol] = []
 
   parsed = await buildout.parse(
@@ -289,29 +310,108 @@ async def lsp_completion(
     ls: LanguageServer,
     params: CompletionParams,
 ) -> Optional[List[CompletionItem]]:
+  await asyncio.sleep(DEBOUNCE_DELAY)
   items: List[CompletionItem] = []
   doc = ls.workspace.get_document(params.textDocument.uri)
 
-  def getInsertText(current_text: str, insert_text: str) -> str:
-    """Calculate the text to be inserted.
-
-    When we already have:
-
-      plone.rec|
-
-    and the completion is `plone.recipe.command`, if we return "plone.recipe.command",
-    this will be appended to the current word, which is `plone`.
-    In this case we want to insert only "recipe.command".
-
-    Note that there is wordPattern in language-configuration.json that is supposed
-    to address this, but it seems we cannot include . in word patterns.
-    There's also the same problem with -
+  def getSectionReferenceCompletionTextEdit(
+      doc: Document,
+      pos: Position,
+      new_text: str,
+  ) -> TextEdit:
+    """Calculate the edition to insert ${section: in ${section:option}
     """
-    for char in ('.', '-'):
-      common_leading, _, _ = current_text.rpartition(char)
-      if common_leading:
-        insert_text = insert_text[len(common_leading) + 1:]
-    return insert_text
+    words_split = re.compile(r'\$\{[-a-zA-Z0-9 ._]*')
+    line = doc.lines[pos.line]
+    index = 0
+    while True:
+      match = words_split.search(line, index)
+      assert match
+      if match.start() <= pos.character <= match.end():
+        start = match.start()
+        end = match.end()
+        return TextEdit(
+            Range(Position(pos.line, start), Position(pos.line, end)),
+            new_text,
+        )
+      index = max(match.start(), index + 1)
+
+    return TextEdit(
+        Range(
+            Position(pos.line, pos.character),
+            Position(pos.line, pos.character),
+        ),
+        new_text,
+    )
+
+  def getOptionReferenceTextEdit(
+      doc: Document,
+      pos: Position,
+      new_text: str,
+  ) -> TextEdit:
+    """Calculate the edition to insert option} a ${section:option}
+    """
+    words_split = re.compile(
+        r'(?P<section>\${[-a-zA-Z0-9 ._]*\:)(?P<option>[-a-zA-Z0-9._]*\}{0,1})')
+    line = doc.lines[pos.line]
+    index = 0
+    while True:
+      match = words_split.search(line, index)
+      assert match
+      section_len = len(match.group('section'))
+      if match.start() + section_len <= pos.character <= match.end():
+        start = match.start() + section_len
+        end = match.end()
+        return TextEdit(
+            Range(Position(pos.line, start), Position(pos.line, end)),
+            new_text,
+        )
+      index = max(match.start(), index + 1)
+
+    return TextEdit(
+        Range(
+            Position(pos.line, pos.character),
+            Position(pos.line, pos.character),
+        ),
+        new_text,
+    )
+
+  def getDefaultTextEdit(
+      doc: Document,
+      pos: Position,
+      new_text: str,
+  ) -> TextEdit:
+    """Calculate the edition to replace the current token at position by the new text.
+    """
+    # regex to split the current token, basically we consider everything a word
+    # but stop at substitution start and end.
+    words_split = re.compile(r'[-a-zA-Z0-9\._\$\{\/]*')
+    line = ''
+    if len(doc.lines) > pos.line:
+      line = doc.lines[pos.line]
+    if not line.strip():
+      return TextEdit(
+          Range(
+              Position(pos.line, pos.character),
+              Position(pos.line, pos.character)),
+          new_text,
+      )
+    index = 0
+    while True:
+      match = words_split.search(line, index)
+      assert match
+      if match.start() <= pos.character <= match.end():
+        start = match.start()
+        end = match.end()
+        # if end was a '}', erase it
+        if (line + '  ')[end] == '}':
+          end += 1
+        # TODO: test
+        return TextEdit(
+            Range(Position(pos.line, start), Position(pos.line, end)),
+            new_text,
+        )
+      index = max(match.start(), index + 1)
 
   parsed = await buildout.open(ls, params.textDocument.uri)
   if parsed is None:
@@ -335,10 +435,12 @@ async def lsp_completion(
         items.append(
             CompletionItem(
                 label=buildout_section_name,
-                insert_text=getInsertText(
-                    symbol.value,
-                    buildout_section_name,
+                text_edit=getSectionReferenceCompletionTextEdit(
+                    doc,
+                    params.position,
+                    '${' + buildout_section_name,
                 ),
+                filter_text='${' + buildout_section_name,
                 kind=CompletionItemKind.Class,
                 documentation=MarkupContent(
                     kind=MarkupKind.Markdown,
@@ -368,7 +470,11 @@ async def lsp_completion(
         items.append(
             CompletionItem(
                 label=buildout_option_name,
-                insert_text=getInsertText(symbol.value, buildout_option_name),
+                text_edit=getOptionReferenceTextEdit(
+                    doc,
+                    params.position,
+                    buildout_option_name + '}',
+                ),
                 kind=CompletionItemKind.Property,
                 documentation=MarkupContent(
                     kind=MarkupKind.Markdown,
@@ -387,7 +493,11 @@ async def lsp_completion(
           items.append(
               CompletionItem(
                   label=option_name,
-                  insert_text=f'{getInsertText(symbol.value, option_name)} = ',
+                  text_edit=getDefaultTextEdit(
+                      doc,
+                      params.position,
+                      option_name + ' = ',
+                  ),
                   kind=CompletionItemKind.Variable,
                   documentation=MarkupContent(
                       kind=MarkupKind.Markdown,
@@ -402,7 +512,11 @@ async def lsp_completion(
           items.append(
               CompletionItem(
                   label=option_name,
-                  insert_text=f'{getInsertText(symbol.value, option_name)} =\n  ',
+                  text_edit=getDefaultTextEdit(
+                      doc,
+                      params.position,
+                      option_name + ' =\n    ',
+                  ),
                   kind=CompletionItemKind.Variable,
                   documentation=MarkupContent(
                       kind=MarkupKind.Markdown,
@@ -416,7 +530,11 @@ async def lsp_completion(
             items.append(
                 CompletionItem(
                     label=k,
-                    insert_text=f'{getInsertText(symbol.value, k)} = ',
+                    text_edit=getDefaultTextEdit(
+                        doc,
+                        params.position,
+                        k + ' = ',
+                    ),
                     kind=CompletionItemKind.Variable,
                     documentation=MarkupContent(
                         kind=MarkupKind.Markdown,
@@ -427,7 +545,11 @@ async def lsp_completion(
           items.append(
               CompletionItem(
                   label='recipe',
-                  insert_text=f'{getInsertText(symbol.value, "recipe")} = ',
+                  text_edit=getDefaultTextEdit(
+                      doc,
+                      params.position,
+                      'recipe = ',
+                  ),
                   kind=CompletionItemKind.Variable))
     elif symbol.kind == buildout.SymbolKind.BuildoutOptionValue:
       # complete option = |
@@ -439,7 +561,8 @@ async def lsp_completion(
           items.append(
               CompletionItem(
                   label=recipe_name,
-                  insert_text=getInsertText(symbol.value, recipe_name),
+                  text_edit=getDefaultTextEdit(doc, params.position,
+                                               recipe_name),
                   kind=CompletionItemKind.Constructor,
                   documentation=MarkupContent(
                       kind=MarkupKind.Markdown,
@@ -452,7 +575,8 @@ async def lsp_completion(
             items.append(
                 CompletionItem(
                     label=section_name,
-                    insert_text=getInsertText(symbol.value, section_name),
+                    text_edit=getDefaultTextEdit(doc, params.position,
+                                                 section_name),
                     kind=CompletionItemKind.Function))
       if symbol.current_section_recipe:
         # complete with recipe options if recipe is known
@@ -462,7 +586,7 @@ async def lsp_completion(
               items.append(
                   CompletionItem(
                       label=valid,
-                      insert_text=getInsertText(symbol.value, valid),
+                      text_edit=getDefaultTextEdit(doc, params.position, valid),
                       kind=CompletionItemKind.Keyword))
       if symbol.current_section_name == 'buildout':
         # complete options of [buildout]
@@ -476,8 +600,9 @@ async def lsp_completion(
             items.append(
                 CompletionItem(
                     label=profile_relative_path,
-                    insert_text=getInsertText(
-                        symbol.value,
+                    text_edit=getDefaultTextEdit(
+                        doc,
+                        params.position,
                         profile_relative_path,
                     ),
                     kind=CompletionItemKind.File,
@@ -492,7 +617,11 @@ async def lsp_completion(
               items.append(
                   CompletionItem(
                       label=section,
-                      insert_text=f'{getInsertText(symbol.value, section)}\n',
+                      text_edit=getDefaultTextEdit(
+                          doc,
+                          params.position,
+                          section + '\n',
+                      ),
                       kind=CompletionItemKind.Function))
 
   return items
@@ -503,6 +632,7 @@ async def lsp_definition(
     ls: LanguageServer,
     params: TextDocumentPositionParams,
 ) -> List[Location]:
+  await asyncio.sleep(DEBOUNCE_DELAY)
   parsed = await buildout.open(ls, params.textDocument.uri)
   if parsed is None:
     return []
@@ -564,12 +694,21 @@ async def lsp_references(
         searched_option = searched_symbol.referenced_option_name
     logger.debug("Looking for references for %s ${%s:%s}", searched_symbol,
                  searched_section, searched_option)
-
+    assert searched_section
     for profile_path in pathlib.Path(
         server.workspace.root_path).glob('**/*.cfg'):
       profile = await buildout.parse(server, profile_path.as_uri())
       if profile is not None:
         assert isinstance(profile, buildout.BuildoutProfile)
+
+        # listing a section in ${buildout:parts} is a reference
+        parts = profile['buildout'].get('parts')
+        if parts is not None and searched_section in parts.value:
+          for option_text, option_range in profile.getOptionValues(
+              'buildout', 'parts'):
+            if searched_section == option_text:
+              references.append(Location(profile.uri, option_range))
+
         async for symbol in profile.getAllOptionReferenceSymbols():
           if symbol.referenced_section_name == searched_section:
             if searched_option is None:
@@ -594,6 +733,7 @@ async def lsp_hover(
     ls: LanguageServer,
     params: TextDocumentPositionParams,
 ) -> Optional[Hover]:
+  await asyncio.sleep(DEBOUNCE_DELAY)
   parsed = await buildout.open(ls, params.textDocument.uri)
   if parsed is None:
     return None
@@ -617,6 +757,7 @@ async def lsp_document_link(
     ls: LanguageServer,
     params: DocumentLinkParams,
 ) -> List[DocumentLink]:
+  await asyncio.sleep(DEBOUNCE_DELAY)
   links: List[DocumentLink] = []
   uri = params.textDocument.uri
   parsed_buildout = await buildout.parse(ls, uri)
